import 'dart:async';
import 'dart:io';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';

import '../types/types.dart';

class SwipeLeftNavigationObserver extends NavigatorObserver {
  ValueNotifier<dynamic> isDidPopRoute;
  String routePath;

  SwipeLeftNavigationObserver(this.routePath, this.isDidPopRoute);

  @override
  void didPop(Route route, Route? previousRoute) {
    // If Patch matched then notify to listener
    RegExp regExp = RegExp(routePath);
    if (regExp.hasMatch(route.settings.name.toString())) {
      isDidPopRoute.value = route;
    }

    super.didPop(route, previousRoute);
  }
}

/// Properties identifying the current state of a [VdoPlayerController].
class VdoPlayerValue {
  /// True if a video is currently being loaded into the player.
  final bool isLoading;

  /// Properties related to the currently loaded video.
  final MediaInfo? mediaInfo;

  /// Total duration of the video.
  ///
  /// This is [Duration.zero] if the video is not loaded.
  final Duration duration;

  /// The current playback position.
  final Duration position;

  /// Position up to where the video is buffered.
  final Duration bufferedPosition;

  /// True if the video is playing, false if paused.
  final bool isPlaying;

  /// True if the video is buffering.
  final bool isBuffering;

  /// True if the currently loaded video has played to the end.
  final bool isEnded;

  /// The current playback speed.
  final double playbackSpeed;

  /// The list of available playback speed options.
  final List<double> playbackSpeedOptions;

  /// True if track selection is done in adaptive mode, false if a track has
  /// been selected explicitly.
  final bool isAdaptive;

  /// Specifies the audio output of the video is muted.
  /// Supported only on web and always false for android and iOS platform.
  final bool isMuted;

  /// Currently selected audio track.
  //final AudioTrack? audioTrack;

  /// Currently selected video track.
  final VideoTrack? videoTrack;

  /// Currently selected subtitle track.
  /// Returns null if subtitle is disabled.
  final SubtitleTrack? subtitleTrack;

  /// List of all available audio tracks.
  //final List<AudioTrack> audioTracks;

  /// List of all available video tracks.
  final List<VideoTrack> videoTracks;

  /// List of all available subtitle tracks.
  final List<SubtitleTrack> subtitleTracks;

  /// Skip Backword/Forward duration
  final Duration skipDuration;

  /// If not null, this gives details about an error encountered while loading
  /// the video or during playback.
  final VdoError? vdoError;

  /// Specifies the current volume of the audio/video. Must be a number between 0.0 and 1.0.
  /// Supported only on web and always null for android and iOS platform.
  final double? volume;

  VdoPlayerValue(
      {this.duration = Duration.zero,
      this.position = Duration.zero,
      this.bufferedPosition = Duration.zero,
      this.isLoading = false,
      this.mediaInfo,
      this.isPlaying = false,
      this.isBuffering = false,
      this.isEnded = false,
      this.playbackSpeed = 1.0,
      this.playbackSpeedOptions = const [],
      this.isAdaptive = true,
      this.isMuted = false,
      this.volume,
      //this.audioTrack,
      this.videoTrack,
      this.subtitleTrack,
      //this.audioTracks = const [],
      this.videoTracks = const [],
      this.subtitleTracks = const [],
      this.skipDuration = const Duration(milliseconds: 10000),
      this.vdoError});

  /// Returns a new instance with the same values as this current instance,
  /// except for any overrides specified as arguments.
  VdoPlayerValue copyWith(
      {bool? isLoading,
      MediaInfo? mediaInfo,
      Duration? duration,
      Duration? position,
      Duration? bufferedPosition,
      bool? isPlaying,
      bool? isBuffering,
      bool? isEnded,
      double? playbackSpeed,
      List<double>? playbackSpeedOptions,
      bool? isAdaptive,
      double? volume,
      bool? isMuted,
      //AudioTrack? audioTrack,
      VideoTrack? videoTrack,
      SubtitleTrack? subtitleTrack,
      //List<AudioTrack>? audioTracks,
      List<VideoTrack>? videoTracks,
      List<SubtitleTrack>? subtitleTracks,
      Duration? skipDuration,
      VdoError? vdoError}) {
    return VdoPlayerValue(
        duration: duration ?? this.duration,
        position: position ?? this.position,
        bufferedPosition: bufferedPosition ?? this.bufferedPosition,
        isLoading: isLoading ?? this.isLoading,
        mediaInfo: mediaInfo ?? this.mediaInfo,
        isPlaying: isPlaying ?? this.isPlaying,
        isBuffering: isBuffering ?? this.isBuffering,
        isEnded: isEnded ?? this.isEnded,
        playbackSpeed: playbackSpeed ?? this.playbackSpeed,
        playbackSpeedOptions: playbackSpeedOptions ?? this.playbackSpeedOptions,
        isAdaptive: isAdaptive ?? this.isAdaptive,
        volume: volume ?? this.volume,
        isMuted: isMuted ?? this.isMuted,
        //audioTrack: audioTrack ?? this.audioTrack,
        videoTrack: videoTrack ?? this.videoTrack,
        subtitleTrack: subtitleTrack ?? this.subtitleTrack,
        //audioTracks: audioTracks ?? this.audioTracks,
        videoTracks: videoTracks ?? this.videoTracks,
        subtitleTracks: subtitleTracks ?? this.subtitleTracks,
        skipDuration: skipDuration ?? this.skipDuration,
        vdoError: vdoError ?? this.vdoError);
  }

  /// Returns a new instance with the same values as this current instance and removes error if it was present.
  VdoPlayerValue copyWithoutError(
      {bool? isLoading,
      MediaInfo? mediaInfo,
      Duration? duration,
      Duration? position,
      Duration? bufferedPosition,
      bool? isPlaying,
      bool? isBuffering,
      bool? isEnded,
      double? playbackSpeed,
      List<double>? playbackSpeedOptions,
      bool? isAdaptive,
      double? volume,
      bool? isMuted,
      //AudioTrack? audioTrack,
      VideoTrack? videoTrack,
      SubtitleTrack? subtitleTrack,
      //List<AudioTrack>? audioTracks,
      List<VideoTrack>? videoTracks,
      List<SubtitleTrack>? subtitleTracks,
      Duration? skipDuration}) {
    return VdoPlayerValue(
        duration: duration ?? this.duration,
        position: position ?? this.position,
        bufferedPosition: bufferedPosition ?? this.bufferedPosition,
        isLoading: isLoading ?? this.isLoading,
        mediaInfo: mediaInfo ?? this.mediaInfo,
        isPlaying: isPlaying ?? this.isPlaying,
        isBuffering: isBuffering ?? this.isBuffering,
        isEnded: isEnded ?? this.isEnded,
        playbackSpeed: playbackSpeed ?? this.playbackSpeed,
        playbackSpeedOptions: playbackSpeedOptions ?? this.playbackSpeedOptions,
        isAdaptive: isAdaptive ?? this.isAdaptive,
        volume: volume ?? this.volume,
        isMuted: isMuted ?? this.isMuted,
        //audioTrack: audioTrack ?? this.audioTrack,
        videoTrack: videoTrack ?? this.videoTrack,
        subtitleTrack: subtitleTrack ?? this.subtitleTrack,
        //audioTracks: audioTracks ?? this.audioTracks,
        videoTracks: videoTracks ?? this.videoTracks,
        subtitleTracks: subtitleTracks ?? this.subtitleTracks,
        skipDuration: skipDuration ?? this.skipDuration,
        vdoError: null);
  }

  //This will allow disabling tracks if its null
  VdoPlayerValue copyWithDisabledTracks(
      {bool? isLoading,
      MediaInfo? mediaInfo,
      Duration? duration,
      Duration? position,
      Duration? bufferedPosition,
      bool? isPlaying,
      bool? isBuffering,
      bool? isEnded,
      double? playbackSpeed,
      List<double>? playbackSpeedOptions,
      bool? isAdaptive,
      double? volume,
      bool? isMuted,

      //AudioTrack? audioTrack,
      VideoTrack? videoTrack,
      SubtitleTrack? subtitleTrack,
      //List<AudioTrack>? audioTracks,
      List<VideoTrack>? videoTracks,
      List<SubtitleTrack>? subtitleTracks,
      Duration? skipDuration,
      VdoError? vdoError}) {
    return VdoPlayerValue(
        duration: duration ?? this.duration,
        position: position ?? this.position,
        bufferedPosition: bufferedPosition ?? this.bufferedPosition,
        isLoading: isLoading ?? this.isLoading,
        mediaInfo: mediaInfo ?? this.mediaInfo,
        isPlaying: isPlaying ?? this.isPlaying,
        isBuffering: isBuffering ?? this.isBuffering,
        isEnded: isEnded ?? this.isEnded,
        playbackSpeed: playbackSpeed ?? this.playbackSpeed,
        playbackSpeedOptions: playbackSpeedOptions ?? this.playbackSpeedOptions,
        isAdaptive: isAdaptive ?? this.isAdaptive,
        volume: volume ?? this.volume,
        isMuted: isMuted ?? this.isMuted,
        //audioTrack: audioTrack,
        videoTrack: videoTrack,
        subtitleTrack: subtitleTrack,
        //audioTracks: audioTracks ?? this.audioTracks,
        videoTracks: videoTracks ?? this.videoTracks,
        subtitleTracks: subtitleTracks ?? this.subtitleTracks,
        skipDuration: skipDuration ?? this.skipDuration,
        vdoError: vdoError ?? this.vdoError);
  }
}

/// Controls a native video player and notifies any listeners of the player
/// state changes.
///
/// The VdoPlayerView widget is used to display the video and it provides
/// an instance of this class to control the video in the onPlayerCreated
/// callback.
class VdoPlayerController extends ValueNotifier<VdoPlayerValue>
    with WidgetsBindingObserver {
  MethodChannel? _channel;
  EventChannel? _eventChannel;
  bool _isDisposed = false;
  AppLifecycleState? appLifecycleCurrentState;

  @protected
  MethodChannel? get channel => _channel;

  @protected
  Function? get funStartPositionTimer => _startPositionTimer;

  @protected
  Function? get funStopPositionTimer => _stopPositionTimer;

  @protected
  set channel(newChannel) {
    _channel = newChannel;
  }

  static final ValueNotifier<dynamic> _didPopRoute =
      ValueNotifier<dynamic>(null);

  Timer? _positionTimer;

  VdoPlayerController() : super(VdoPlayerValue());

  VdoPlayerController.withChannels(this._channel, this._eventChannel)
      : super(VdoPlayerValue());

  void init() {
    _eventChannel
        ?.receiveBroadcastStream()
        .listen(_handleNativeEvent, onError: _handleNativeErrorEvent);
    WidgetsBinding.instance.addObserver(this);

    // When Swipe left event triggered on iOS
    _didPopRoute.addListener(() {
      _packUpPlayback();
    });
  }

  static navigatorObserver(String routePathOrRegExp) {
    return SwipeLeftNavigationObserver(routePathOrRegExp, _didPopRoute);
  }

  @override
  void dispose() {
    if (!kIsWeb) {
      if (Platform.isIOS) {
        _channel?.invokeMethod("dispose");
      }
    }
    _stopPositionTimer();
    WidgetsBinding.instance.removeObserver(this);
    if (!_isDisposed) {
      _isDisposed = true;
      super.dispose();
    }
  }

  @override
  void notifyListeners() {
    //Do not notify if controller is disposed.
    if (!_isDisposed) super.notifyListeners();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    super.didChangeAppLifecycleState(state);
    // TODO didChangeAppLifecycleState often called multiple times, fix!
    if (state == AppLifecycleState.paused) {
      appLifecycleCurrentState = state;
      _packUpPlayback();
    }
    if (state == AppLifecycleState.resumed) {
      // Restore with playback state saved on native side
      if (appLifecycleCurrentState == AppLifecycleState.paused) _restore();
    }
  }

  void _packUpPlayback() {
    _stopPositionTimer();
    _packUp(); // Save playback state and release player resources
  }

  bool isDisposed() {
    return _isDisposed;
  }

  bool get isPlaying => value.isPlaying;

  bool get isBuffering => value.isBuffering;

  bool get isLoading => value.isLoading;

  bool get isEnded => value.isEnded;

  /// Load a new video with specified options to the player as identified by the
  /// embedInfo argument.
  Future<void> load(EmbedInfo embedInfo) async {
    try {
      /// Throw an Exception if Playback speed is less than 0.25x and greater than 2.25x
      embedInfo.playbackSpeedOptions?.forEach((i) {
        if (i < 0.25 || i > 2.25) {
          throw Exception('Playback speed allowed between 0.25x to 2.25x');
        }
      });

      // Renew VdoPlayerValue, remove old video related parameters
      _stopPositionTimer();
      value = VdoPlayerValue(isLoading: true);
      notifyListeners();

      Map<String, Object> loadArgs = {
        'restore': false,
        'embedInfo': embedInfo.toMap()
      };
      Map loadResultMap = await _channel!.invokeMethod("load", loadArgs);
      Map mediaInfoMap = loadResultMap['mediaInfo'];

      MediaInfo mediaInfo = MediaInfo(
          mediaId: mediaInfoMap['mediaId'],
          type: MediaInfoType.streaming,
          // TODO fix
          title: mediaInfoMap['title'],
          description: mediaInfoMap['description'],
          duration: Duration(milliseconds: mediaInfoMap['duration']));

      value = value.copyWith(
          isLoading: false,
          mediaInfo: mediaInfo,
          duration: mediaInfo.duration,
          isPlaying: loadResultMap['playWhenReady'],
          skipDuration: embedInfo.skipDuration,
          playbackSpeedOptions: embedInfo.playbackSpeedOptions);

      notifyListeners();

      if (value.isPlaying) {
        _startPositionTimer();
      }
    } on PlatformException catch (e) {
      debugPrint("load error [ ${e.code} , ${e.message!} + ]");
      String errMsg = e.message ?? "unknown error";
      VdoError error = VdoError(int.parse(e.code), errMsg, null);
      value =
          value.copyWith(isLoading: false, isBuffering: false, vdoError: error);
      notifyListeners();
    }
  }

  /// Start or resume playing the video.
  Future<void> play() async {
    value = value.copyWith(isPlaying: true);
    _startPositionTimer();
    await _channel!.invokeMethod("setPlayWhenReady", isPlaying);
    notifyListeners();
  }

  /// Pause video playback.
  Future<void> pause() async {
    value = value.copyWith(isPlaying: false);
    _stopPositionTimer();
    await _channel!.invokeMethod("setPlayWhenReady", isPlaying);
    notifyListeners();
  }

  /// Unload the currently loaded video from the player.
  Future<void> stop() async {
    _stopPositionTimer();
    await _channel!.invokeMethod("stop");
    notifyListeners();
  }

  /// Returns the total duration of the video.
  Future<Duration> getDuration() async {
    int durationMs =
        await (_channel!.invokeMethod("getDuration") as FutureOr<int>);
    return Duration(milliseconds: durationMs);
  }

  /// Returns an estimate of the position up to which data is buffered.
  Future<Duration> getBufferedPosition() async {
    List positions = await _getPlaybackAndBufferedPositions();
    return positions[1];
  }

  /// Returns the current playback position.
  Future<Duration> getPosition() async {
    List positions = await _getPlaybackAndBufferedPositions();
    return positions[0];
  }

  /// Sets the volume of the audio/video. Must be a number between 0.0 and 1.0
  Future<void> setVolume(double volume) async {
    throw UnimplementedError(
        'setVolume() has not been implemented for iOS and Android.');
  }

  /// Mutes or unmutes the player based on the values passed
  Future<void> setMute(bool mute) async {
    throw UnimplementedError(
        'setMute() has not been implemented for iOS and Android.');
  }

  /// Seek to the specified target position.
  Future<void> seek(Duration target) async {
    target = target > value.duration ? value.duration : target;
    await _channel!.invokeMethod("seek", target.inMilliseconds);
    notifyListeners();
  }

  /// Set the video playback speed.
  Future<void> setPlaybackSpeed(double speed) async {
    value = value.copyWith(playbackSpeed: speed);
    await _channel!.invokeMethod("setPlaybackSpeed", speed);
    notifyListeners();
  }

  Future<void> setAdaptive() async {
    await _channel!.invokeMethod("setAdaptive");
    value = value.copyWith(isAdaptive: true);
    notifyListeners();
  }

  /// Returns values of additional properties specific to current playback session. Returned values
  /// need to be cast to their expected types before use.
  ///
  /// For e.g. value for property "totalPlayed" is total played duration in seconds of type [int]
  ///
  /// [propertyName] name of desired property
  ///
  /// Returns requested property value as an [Object?]

  Future<Object?> getPlaybackProperty(String propertyName) async {
    Object? playbackProperty =
        await _channel!.invokeMethod("getPlaybackProperty", propertyName);
    return Future.value(playbackProperty);
  }

  Future<void> setVideoTrack(VideoTrack videoTrack) async {
    await _channel!.invokeMethod("setVideoTrack", videoTrack.id);
    value = value.copyWith(isAdaptive: false);
    notifyListeners();
  }

  /// Set subtitle language. Pass null to disable subtitles.
  Future<void> setSubtitleLanguage(String? language) async {
    bool changed =
        await _channel!.invokeMethod("setSubtitleLanguage", language);
    if (!changed) {
      debugPrint('Subtitle language not changed.');
    } else if (Platform.isIOS) {
      _setSubtitle(language);
    }
    notifyListeners();
  }

  _startPositionTimer() {
    if (_positionTimer != null) {
      _positionTimer!.cancel();
    }
    _positionTimer =
        Timer.periodic(const Duration(milliseconds: 500), (timer) async {
      _updatePlayerPosition();
    });
  }

  _updatePlayerPosition() async {
    Duration position = await getPosition();
    Duration bufferedPosition = await getBufferedPosition();
    value =
        value.copyWith(position: position, bufferedPosition: bufferedPosition);
  }

  _stopPositionTimer() {
    if (_positionTimer != null) {
      _positionTimer!.cancel();
    }
  }

  Future<List<Duration>> _getPlaybackAndBufferedPositions() async {
    Map positions =
        await _channel!.invokeMethod("getPlaybackAndBufferedPositions");
    int positionMs = positions['position'];
    int bufferedPositionMs = positions['bufferedPosition'];
    return [
      Duration(milliseconds: positionMs),
      Duration(milliseconds: bufferedPositionMs)
    ];
  }

  Future<void> setBackPressCalled() async {
    await _channel!.invokeMethod("onBackPressed");
  }

  Future<void> _packUp() async {
    await _channel!.invokeMethod("packUp");
  }

  Future<void> _restore() async {
    //Not creating new instance for player value to maintain player state on restore.
    //value = VdoPlayerValue(isLoading: true);
    value = value.copyWith(isLoading: true);
    Map<String, Object> loadArgs = {
      'restore': true,
      'lastPlaybackSpeed': value.playbackSpeed,
    };
    Map loadResultMap = await _channel!.invokeMethod("load", loadArgs);

    bool didRestore = loadResultMap['didLoad'];
    if (!didRestore) {
    } else {
      Map mediaInfoMap = loadResultMap['mediaInfo'];

      MediaInfo mediaInfo = MediaInfo(
          mediaId: mediaInfoMap['mediaId'],
          type: MediaInfoType.streaming,
          // TODO fix
          title: mediaInfoMap['title'],
          description: mediaInfoMap['description'],
          duration: Duration(milliseconds: mediaInfoMap['duration']));

      value = value.copyWith(
          isLoading: false,
          mediaInfo: mediaInfo,
          duration: mediaInfo.duration,
          isPlaying: loadResultMap['playWhenReady']);

      if (value.isPlaying) {
        _startPositionTimer();
      }
    }
  }

  void _updateTracks(List availableTracks, List selectedTracks) {
    final List<AudioTrack> allAudio = [];
    final List<VideoTrack> allVideo = [];
    final List<SubtitleTrack> allSubtitle = [];

    for (var element in availableTracks) {
      if (element is Map) {
        switch (element["type"]) {
          case "audio":
            allAudio.add(AudioTrack(
                id: element["id"],
                language: element["language"],
                bitrate: element["bitrate"]));
            break;
          case "video":
            allVideo.add(VideoTrack(
                id: element["id"],
                bitrate: element["bitrate"],
                width: element["width"],
                height: element["height"]));
            break;
          case "subtitle":
            allSubtitle.add(SubtitleTrack(
                id: element["id"], language: element["language"]));
            break;
        }
      }
    }

    //AudioTrack? selectedAudio;
    VideoTrack? selectedVideoTrack;
    SubtitleTrack? selectedSubtitle;
    //If there is no subtitle track in selected tracks set selected subtitle to null
    selectedSubtitle = null;
    //If there is no video track available in selected tracks set selected video track to null
    selectedVideoTrack = null;
    int selectedVideoTracksCount = 0;
    for (var element in selectedTracks) {
      if (element is Map) {
        switch (element["type"]) {
          case "audio":
            /*selectedAudio = AudioTrack(
                id: element["id"],
                language: element["language"],
                bitrate: element["bitrate"]);*/
            break;
          case "video":
            selectedVideoTracksCount++;
            selectedVideoTrack = VideoTrack(
                id: element["id"],
                bitrate: element["bitrate"],
                width: element["width"],
                height: element["height"]);
            break;
          case "subtitle":
            selectedSubtitle =
                SubtitleTrack(id: element["id"], language: element["language"]);
            break;
        }
      }
    }

    //Set selected video track as null if list of selected tracks have more than one video track.
    if (selectedVideoTracksCount > 1) {
      selectedVideoTrack = null;
    }

    value = value.copyWith(
        //audioTrack: selectedAudio,
        videoTrack: selectedVideoTrack,
        subtitleTrack: selectedSubtitle,
        //audioTracks: allAudio,
        videoTracks: allVideo,
        subtitleTracks: allSubtitle);

    //This will allow disabling subtitle if its null
    value = value.copyWithDisabledTracks(
        subtitleTrack: selectedSubtitle, videoTrack: selectedVideoTrack);
  }

  void _handleNativeEvent(nativeEvent) {
    if (nativeEvent is Map) {
      Map<dynamic, dynamic> event = nativeEvent;
      switch (event["name"]) {
        case "onPlayerStateChanged":
          final String? playbackState = event["playbackState"];
          final bool ended = playbackState == 'ended';
          final bool ready = playbackState == 'ready';
          if (ended) {
            _stopPositionTimer();
          } else if (ready) {
            //start position timer if playback state is ready and timer is not active.
            if (_positionTimer == null) {
              _startPositionTimer();
            } else if (!_positionTimer!.isActive) {
              _startPositionTimer();
            }
          }
          // Set isEnded value here instead of in onMediaEnded, because it updates
          // if the playback is resumed after end
          value = value.copyWith(
              isBuffering: playbackState == 'buffering',
              isEnded: ended,
              isPlaying: event['playWhenReady'],
              position: ended ? value.duration : value.position);
          break;
        case "onPlaybackPositionChanged":
          //Only called from iOS side
          _updatePlayerPosition();
          break;
        case "onPlaybackSpeedChanged":
          final double playbackSpeed = event["value"];
          value = value.copyWith(playbackSpeed: playbackSpeed);
          break;
        case "onMediaEnded":
          break;
        case "onTracksChanged":
          final List availableTracks = event["available"];
          final List selectedTracks = event["selected"];
          _updateTracks(availableTracks, selectedTracks);
          break;
        case "playbackSessionInterrupted":
          _packUpPlayback();
          break;
        case "onError":
          Map<Object?, Object?> errorDescription = event['errorDescription'];
          VdoError vdoError = VdoError.fromMap(errorDescription);
          value = value.copyWith(vdoError: vdoError);
          break;
        default:
          break;
      }
    }
  }

  void _setSubtitle(String? langCode) {
    SubtitleTrack? selectedSubtitleTrack;
    for (SubtitleTrack subtitleTrack in value.subtitleTracks) {
      if (subtitleTrack.language == langCode) {
        selectedSubtitleTrack = subtitleTrack;
      }
    }
    value = value.copyWithDisabledTracks(
        subtitleTrack: selectedSubtitleTrack, videoTrack: value.videoTrack);
  }

  void _handleNativeErrorEvent(nativeError) {
    if (nativeError is PlatformException) {
      PlatformException error = nativeError;
      if (error.code.isNotEmpty) {
        debugPrint(
            'Native player error event: ${error.code}, ${error.message!}');
      } else {
        debugPrint('Native player error event: ${error.message!}');
      }
      _stopPositionTimer();
      String errMsg = error.message ?? "Unknown error";
      value = value.copyWith(
          isLoading: false,
          isBuffering: false,
          vdoError: VdoError(error.code.isNotEmpty ? int.parse(error.code) : -1,
              errMsg, null));
    }
  }
}
