import 'package:flutter/widgets.dart';
import 'package:plugin_platform_interface/plugin_platform_interface.dart';
import 'package:vdocipher_flutter/vdocipher_flutter.dart';

/// The interface that implementations of vdo_web_player must implement.
///
abstract class VideoPlayerPlatform extends PlatformInterface {
  /// Constructs a VideoPlayerPlatform.
  VideoPlayerPlatform() : super(token: _token);

  static final Object _token = Object();

  static VideoPlayerPlatform _instance = _PlaceholderImplementation();

  /// The instance of [VideoPlayerPlatform] to use.
  ///
  /// Defaults to a placeholder that does not override any methods, and thus
  /// throws `UnimplementedError` in most cases.
  static VideoPlayerPlatform get instance => _instance;

  /// Platform-specific plugins should override this with their own
  /// platform-specific class that extends [VideoPlayerPlatform] when they
  /// register themselves.
  static set instance(VideoPlayerPlatform instance) {
    PlatformInterface.verify(instance, _token);
    _instance = instance;
  }

  /// Initializes the platform interface and disposes all existing players.
  ///
  /// This method is called when the plugin is first initialized
  /// and on every full restart.
  Future<void> init() async {
    throw UnimplementedError('init() has not been implemented.');
  }

  /// Clears one video.
  Future<void> dispose(int textureId) {
    throw UnimplementedError('dispose() has not been implemented.');
  }

  /// Creates an instance of a video player and returns its textureId.
  Future<int?> create(EmbedInfo embedInfo, bool showControls) {
    throw UnimplementedError('create() has not been implemented.');
  }

  /// Sets the looping attribute of the video.
  Future<void> setLooping(int textureId, bool looping) {
    throw UnimplementedError('setLooping() has not been implemented.');
  }

  /// Starts the video playback.
  Future<void> play(int textureId) {
    throw UnimplementedError('play() has not been implemented.');
  }

  Future<void> updatePlayerConfigMap(int textureId, Map? configMap) async {
    throw UnimplementedError(
        'updatePlayerConfigMap() has not been implemented.');
  }

  /// Load a video for playback.
  Future<void> load(int textureId, String otp, String playbackInfo) {
    throw UnimplementedError('load() has not been implemented.');
  }

  /// Stops the video playback.
  Future<void> pause(int textureId) {
    throw UnimplementedError('pause() has not been implemented.');
  }

  /// Sets the volume to a range between 0.0 and 1.0.
  Future<void> setVolume(int textureId, double volume) async {
    throw UnimplementedError('setVolume() has not been implemented.');
  }

  /// Mutes or unmutes the player based on the values passed.
  Future<void> setMute(int textureId, bool mute) async {
    throw UnimplementedError('setMute() has not been implemented.');
  }

  /// Returns player the volume to a range between 0.0 and 1.0.
  Future<double> getVolume(int textureId) async {
    throw UnimplementedError('getVolume() has not been implemented.');
  }

  /// Sets the video position to a [Duration] from the start.
  Future<void> seekTo(int textureId, Duration position) {
    throw UnimplementedError('seekTo() has not been implemented.');
  }

  /// Sets the playback speed to a [speed] value indicating the playback rate.
  Future<void> setPlaybackSpeed(int textureId, double speed) {
    throw UnimplementedError('setPlaybackSpeed() has not been implemented.');
  }

  /// Gets the video position as [Duration] from the start.
  Future<Duration> getPosition(int textureId) {
    throw UnimplementedError('getPosition() has not been implemented.');
  }

  /// Gets the video buffered position as [Duration] from the start.
  Future<Duration> getBufferedPosition(int textureId) {
    throw UnimplementedError('getBufferedPosition() has not been implemented.');
  }

  Future<VideoTrack?> getVideoTrack(int textureId) async {
    throw UnimplementedError('getVideoTrack() has not been implemented.');
  }

  Future<SubtitleTrack?> getSubtitleTrack(int textureId) async {
    throw UnimplementedError('getSubtitleTrack() has not been implemented.');
  }

  Future<List<VideoTrack>> getVideoTracks(int textureId) async {
    throw UnimplementedError('getVideoTracks() has not been implemented.');
  }

  Future<List<SubtitleTrack>> getSubtitleTracks(int textureId) async {
    throw UnimplementedError('getSubtitleTracks() has not been implemented.');
  }

  /// Gets the video position as [Duration] from the start.
  Future<Duration> getDuration(int textureId) {
    throw UnimplementedError('getDuration() has not been implemented.');
  }

  Future<void> setAdaptive(int textureId) {
    throw UnimplementedError('setAdaptive() has not been implemented.');
  }

  Future<void> setVideoTrackId(int textureId, int videoTrackId) {
    throw UnimplementedError('setVideoTrackId() has not been implemented.');
  }

  Future<void> setTextTrackLanguage(int textureId, String? language) {
    throw UnimplementedError(
        'setTextTrackLanguage() has not been implemented.');
  }

  Future<int> getPlaybackProperty(int textureId, String propertyName) async {
    throw UnimplementedError('getPlaybackProperty() has not been implemented.');
  }

  Future<MediaInfo> getMediaInfo(int textureId, String mediaId) async {
    throw UnimplementedError('getMediaInfo() has not been implemented.');
  }

  /// Returns a widget displaying the video with a given textureID.
  Widget buildView(int textureId) {
    throw UnimplementedError('buildView() has not been implemented.');
  }

  /// Sets the audio mode to mix with other sources
  Future<void> setMixWithOthers(bool mixWithOthers) {
    throw UnimplementedError('setMixWithOthers() has not been implemented.');
  }

  /// Returns a Stream of [VideoEventType]s.
  Stream<VideoEvent> videoEventsFor(int textureId) {
    throw UnimplementedError('videoEventsFor() has not been implemented.');
  }
}

class Error {
  String? message;
  int? code;

  Error({this.message, this.code});
}

class VideoEvent {
  VideoEvent(
      {required this.eventType,
      this.duration,
      this.progress,
      this.playbackRate});

  VideoEvent copyWith(
      {VideoEventType? eventType,
      Duration? duration,
      Duration? progress,
      Duration? bufferedPosition,
      Error? error,
      double? playbackRate,
      int? videoTrackId,
      int? textTrackId,
      double? volume,
      bool? isAdaptive,
      bool? isMuted,
      bool? isCaptionEnabled}) {
    if (eventType != null) {
      this.eventType = eventType;
    }
    this.duration = duration ?? this.duration;
    this.progress = progress ?? this.progress;
    this.error = error ?? this.error;
    this.bufferedPosition = bufferedPosition ?? this.bufferedPosition;
    this.playbackRate = playbackRate ?? this.playbackRate;
    this.videoTrackId = videoTrackId ?? this.videoTrackId;
    this.textTrackId = textTrackId ?? this.textTrackId;
    this.isAdaptive = isAdaptive ?? this.isAdaptive;
    this.volume = volume ?? this.volume;
    this.isMuted = isMuted ?? this.isMuted;
    this.isCaptionEnabled = isCaptionEnabled ?? this.isCaptionEnabled;
    return this;
  }

  /// The type of the event.
  VideoEventType eventType;

  /// Duration of the video.
  Duration? duration;

  Error? error;

  Duration? progress;

  Duration? bufferedPosition;

  double? playbackRate;

  double? volume;

  int? videoTrackId;

  int? textTrackId;

  bool isAdaptive = false;

  bool isCaptionEnabled = false;

  bool isMuted = false;
}

/// Type of the event.
///
/// Emitted by the platform implementation when the video is initialized or
/// completed or to communicate buffering events or play state changed.
enum VideoEventType {
  /// The video has been initialized.
  initialized,
  apiReady,
  waiting,
  play,
  playing,
  pause,
  loadstart,
  loadeddata,
  error,

  /// The playback has ended.
  ended,
  ratechange,
  progress,

  /// Updated information on the buffering state.
  bufferingUpdate,

  /// The video started to buffer.
  bufferingStart,

  /// The video stopped to buffer.
  bufferingEnd,

  /// Video quality changed.
  videoQualityChange,

  /// Adaptive playback changed.
  videoAdaptivenessChange,

  /// Caption language changed.
  captionLanguageChange,

  /// Caption visibility changed.
  captionVisibilityChange,

  /// Volume change
  volumechange,

  /// An unknown event has been received.
  unknown,
}

class _PlaceholderImplementation extends VideoPlayerPlatform {}
