import 'dart:async';
import 'dart:collection';
import 'dart:convert';
import 'dart:html' as html;
import 'dart:js_util';

import 'package:flutter/material.dart';
import 'package:flutter_web_plugins/flutter_web_plugins.dart';
import 'package:js/js.dart' as js;
import 'package:vdocipher_flutter/src/types/media_info.dart';
import 'package:vdocipher_flutter/src/types/track/subtitle_track.dart';
import 'package:vdocipher_flutter/src/types/track/video_track.dart';

import '../types/embed_info.dart';
import 'shims/dart_ui.dart' as ui;
import 'video_player.dart';
import 'vdo_player_platform_interface.dart';

@js.JS()
external dynamic updateConfigMap(int textureId, String configMap);

@js.JS()
external dynamic getVideoQualities(int textureId);

@js.JS()
external dynamic getCaptionLanguages(int textureId);

@js.JS()
external dynamic fetchPlaybackProperty(int textureId, String propertyName);

@js.JS()
external dynamic setAdaptivePlayback(int textureId);

@js.JS()
external dynamic setVideoQuality(int textureId, int videoTrackId);

@js.JS()
external dynamic setTextTrack(int textureId, String language);

@js.JS()
external dynamic disableSubtitle(int textureId);

@js.JS()
external dynamic getPlayerVolume(int textureId);

@js.JS()
external dynamic setPlayerVolume(int textureId, double volume);

@js.JS()
external dynamic setPlayerMute(int textureId, bool mute);

@js.JS()
external dynamic loadVideo(int textureId, String otp, String playbackInfo);

@js.JS()
external dynamic playVideo(int textureId);

@js.JS()
external dynamic pauseVideo(int textureId);

@js.JS()
external dynamic getVideoDuration(int textureId);

@js.JS()
external dynamic getVideoPosition(int textureId);

@js.JS()
external dynamic getVideoBufferedPosition(int textureId);

@js.JS()
external dynamic setPlaybackRate(int textureId, double playbackRate);

@js.JS()
external dynamic videoSeekTo(int textureId, int position);

@js.JS()
external dynamic getMetaData(int textureId);

class ItemUrlPolicy implements html.UriPolicy {
  RegExp regex = RegExp(r'(?:http://|https://)?.*');

  @override
  bool allowsUri(String uri) {
    return regex.hasMatch(uri);
  }
}

/// The web implementation of [VideoPlayerPlatform].
///
/// This class implements the `package:video_player` functionality for the web.
class VdoCipherPlayerPlugin extends VideoPlayerPlatform {
  /// Registers this class as the default instance of [VideoPlayerPlatform].
  static void registerWith(Registrar registrar) {
    VideoPlayerPlatform.instance = VdoCipherPlayerPlugin();
  }

  // Map of textureId -> VideoPlayer instances
  final Map<int, VideoPlayer> _videoPlayers = <int, VideoPlayer>{};

  // Simulate the native "textureId".
  int _textureCounter = 1;

  @override
  Future<void> init() async {
    return _disposeAllPlayers();
  }

  @override
  Future<void> dispose(int textureId) async {
    _player(textureId).dispose();
    _videoPlayers.remove(textureId);
    return;
  }

  void _disposeAllPlayers() {
    for (final VideoPlayer videoPlayer in _videoPlayers.values) {
      videoPlayer.dispose();
    }
    _videoPlayers.clear();
  }

  @override
  Future<int> create(EmbedInfo embedInfo, bool showControls) async {
    final int textureId = _textureCounter++;

    String url = "https://player.vdocipher.com/v2/player.js";

    var isCustomElementExists =
        html.window.customElements?.get('vdocipher-player') != null;

    html.BodyElement htmlBodyElement = html.BodyElement();

    if (!isCustomElementExists) {
      html.ScriptElement scriptElement = html.ScriptElement()..src = url;
      htmlBodyElement.append(scriptElement);
    }

    String controls =
        embedInfo.embedInfoOptions?.nativeControls == true ? "native" : "on";

    String? enforceiframe = "true";
    if (!showControls) {
      controls = "off";
      enforceiframe = null;
    }

    String autoplay =
        embedInfo.embedInfoOptions?.autoplay == true ? "true" : "false";

    String customPlayerId = embedInfo.embedInfoOptions?.customPlayerId ?? '';

    String playbackInfo = embedInfo.playbackInfo ?? '';
    String otp = embedInfo.otp ?? '';
    String token = embedInfo.token ?? '';
    String videoId = embedInfo.mediaId ?? '';

    String ccLanguage =
        embedInfo.embedInfoOptions?.preferredCaptionsLanguage ?? '';
    htmlBodyElement = htmlBodyElement
      ..appendHtml("""
              <vdocipher-player
                id="vdo_player_$textureId"
                otp="$otp"
                token="$token"
                videoId="$videoId"
                player='$customPlayerId'
                enforceiframe='$enforceiframe'
                playbackInfo="$playbackInfo"
                ccLanguage=$ccLanguage
                autoplay=$autoplay
                controls=$controls
                style="width: 100%; height: 100%;"
              >
              </vdocipher-player>
            <script defer>
            
            var vdoPlayerElement = document.querySelector("#vdo_player_" + $textureId);
            
            if(vdoPlayerElement.player) {
              playerEventListeners.addEventListener([$textureId]);
              sendEvent(createEventObject($textureId, 'apiReady'));
            } else {
              vdoPlayerElement.addEventListener('apiReady', () => {
                 playerEventListeners.addEventListener([$textureId]);
                 sendEvent(createEventObject($textureId, 'apiReady'));
              })
            }

  function isElementRegistered() {
    return customElements.get('vdocipher-player')!==undefined;
  }
  
  function updateConfigMap(textureId, configMap) {
      const playerComponent = document.querySelector('#vdo_player_' + textureId);
      return playerComponent.player.api.updatePlayerConfig(JSON.parse(configMap));
  }
  
  function getVideoQualities(textureId) {

      const playerComponent = document.querySelector('#vdo_player_' + textureId);
      return playerComponent.player.api.getVideoQualities();
  }

  function getCaptionLanguages(textureId) {
      const playerComponent = document.querySelector('#vdo_player_' + textureId);
      return playerComponent.player.api.getCaptionLanguages();
  }

  function fetchPlaybackProperty(textureId, propertyName) {
      const playerComponent = document.querySelector('#vdo_player_' + textureId);
      if(propertyName == 'totalPlayed')
      {
         return playerComponent.player.api.getTotalPlayed();
      }
      else if(propertyName == 'totalCovered')
      {
         return playerComponent.player.api.getTotalCovered();
      } else {
         return new Promise(-1);
      }
  }

  function setAdaptivePlayback(textureId) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       return playerComponent.player.api.enableAdaptiveVideo();
  }

  function setVideoQuality(textureId, videoTrackId) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       return playerComponent.player.api.setVideoQuality(videoTrackId);
  }

  function setTextTrack(textureId, language) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       return playerComponent.player.api.setCaptionLanguage(language);
  }

  function disableSubtitle(textureId) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       return playerComponent.player.api.hideCaptions();
  }

  function getPlayerVolume(textureId) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       return playerComponent.player.video.volume;
  }

  function setPlayerVolume(textureId, volume) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       playerComponent.player.video.volume = volume;
  }
  
  function setPlayerMute(textureId, mute) {
       const playerComponent = document.querySelector('#vdo_player_' + textureId);
       playerComponent.player.video.muted = mute;
  }

  function loadVideo(textureId, otp, playbackInfo) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    playerComponent.player.api.loadVideo({'otp': otp, 'playbackInfo': playbackInfo});
  }
  
  function playVideo(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    return playerComponent.player.video.play()
  }
  
  function pauseVideo(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    return playerComponent.player.video.pause();
  }
  
  function setPlaybackRate(textureId, playbackRate) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    playerComponent.player.video.playbackRate = playbackRate;
  }
  
  function getVideoDuration(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    if(playerComponent.player == null) return 0;
    return playerComponent.player.video.duration;
  }
  
  function getVideoPosition(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    if(playerComponent.player == null) return 0;
    return playerComponent.player.video.currentTime;
  }
  
  function getMetaData(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    if(playerComponent.player == null) return Promise.resolve({});
    return playerComponent.player.api.getMetaData();
  }
  
  function getVideoBufferedPosition(textureId) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    if(playerComponent.player == null) return 0;
    if(playerComponent.player.video.buffered.__store === undefined) return 0;
    const length = playerComponent.player.video.buffered.__store.length;
    let bufferedTime = 0;
    if(length > 0) {
        let currentTime = playerComponent.player.video.currentTime;
        bufferedTime = playerComponent.player.video.buffered.end(length-1);
        bufferedTime = Math.max(currentTime, bufferedTime);
    }
    return bufferedTime;
  }
  
  function videoSeekTo(textureId, position) {
    const playerComponent = document.querySelector('#vdo_player_' + textureId);
    playerComponent.player.video.currentTime = position;
  }

  function createEventObject(textureId, name, value, muted) {
      const payload = {
         'textureId': textureId,
         'name': name,
         'value': value,
         'muted': muted,
      };
      return payload;
  }
  
  function createErrorObject(textureId, name, errorObj) {
     let errorCode = 0;
     if('code' in errorObj)
       {
         if(isNaN(errorObj.code) === false){
           errorCode = parseInt(errorObj.code);
         }
         else {
           errorCode = errorObj.code;
         }
       }
     const message = errorObj.message;
     
     const payload = {
         'textureId': textureId,
         'name': name,
         'value': message,
         'code': errorCode,
      };
     return payload;
  }

  function sendEvent(data) {
     const jsonString = JSON.stringify(data);
     onVdoPlayerEvent(jsonString);
  }

  var playerEventListeners = {
  
   addEventListener: function(arguments) {

   const textureId = arguments[0];

   const playerComponent = document.querySelector('#vdo_player_' + textureId);

   sendEvent(createEventObject(textureId, 'initialized'));

       playerComponent.player.video.addEventListener('play', function(event){
                    sendEvent(createEventObject(textureId, 'play'));
                });
       playerComponent.player.video.addEventListener('pause', function(event){
                    sendEvent(createEventObject(textureId, 'pause'));
                });
        playerComponent.player.video.addEventListener('waiting', function(event){
                    sendEvent(createEventObject(textureId, 'waiting'));
                });
       playerComponent.player.video.addEventListener('playing', function(event){
                    sendEvent(createEventObject(textureId, 'playing'));
                });
       playerComponent.player.video.addEventListener('progress', function(event){
                    if(playerComponent.player.video.buffered){
                    const length = playerComponent.player.video.buffered.__store.length;
                    let bufferedTime = 0;
                    if(length > 0) {
                       const currentTime = playerComponent.player.video.currentTime;
                       bufferedTime = playerComponent.player.video.buffered.end(length-1);
                       bufferedTime = Math.max(currentTime, bufferedTime);
                    }
                    sendEvent(createEventObject(textureId, 'buffered', bufferedTime));
             }
                });         
       playerComponent.player.video.addEventListener('progress', function(event){
                    const currentTime = playerComponent.player.video.currentTime;
                    sendEvent(createEventObject(textureId, 'progress', currentTime));
                });
       playerComponent.player.video.addEventListener('ratechange', function(event){
                    const playbackRate = playerComponent.player.video.playbackRate;
                    sendEvent(createEventObject(textureId, 'ratechange', playbackRate));
                });
       playerComponent.player.video.addEventListener('ended', function(event){
                    sendEvent(createEventObject(textureId, 'ended'));
                });
        playerComponent.player.video.addEventListener('loadstart', function(event){
                    sendEvent(createEventObject(textureId, 'loadstart'));
                });
        playerComponent.player.video.addEventListener('loadeddata', function(event){
                    const duration = playerComponent.player.video.duration;
                    sendEvent(createEventObject(textureId, 'loadeddata', duration));
                });
        playerComponent.player.video.addEventListener('volumechange', function(event){
                    const volume = playerComponent.player.video.volume;
                    const muted = playerComponent.player.video.muted;
                    sendEvent(createEventObject(textureId, 'volumechange', volume, muted));
                });
       playerComponent.player.api.addEventListener('videoQualityChange', function(trackId) {
                    sendEvent(createEventObject(textureId, 'videoQualityChange', parseInt(trackId)));
                });
       playerComponent.player.api.addEventListener('videoAdaptivenessChange', function(value) {
                    sendEvent(createEventObject(textureId, 'videoAdaptivenessChange', value));
                });
       playerComponent.player.api.addEventListener('captionLanguageChange', function(param) {
                    if (param === 'TURN_OFF') return;
                    sendEvent(createEventObject(textureId, 'captionLanguageChange', parseInt(param)));
                });
       playerComponent.player.api.addEventListener('captionVisibilityChange', function(param) {
                    sendEvent(createEventObject(textureId, 'captionVisibilityChange', param));
                });
       playerComponent.player.api.addEventListener('error', function(param) {
                    sendEvent(createErrorObject(textureId, 'error', param));
                });         
   }
};
            </script>
          """,
          validator: html.NodeValidatorBuilder()
            ..allowCustomElement('vdocipher-player', attributes: [
              'playbackinfo',
              'otp',
              'token',
              'videoId',
              'style',
              'enforceiframe',
              'player',
              'controls',
              'cclanguage',
              'autoplay'
            ])
            ..allowCustomElement('script', attributes: ['src', 'defer'])
            ..allowCustomElement('div', attributes: ['style'])
            ..allowHtml5(uriPolicy: ItemUrlPolicy()))
      ..style.border = 'none'
      ..style.margin = '0px'
      ..style.height = '100%'
      ..style.width = '100%';

    try {
      ui.platformViewRegistry.registerViewFactory(
          "vdoCipherPlayer-$textureId", (int viewId) => htmlBodyElement);
      final VideoPlayer player = VideoPlayer(videoElement: htmlBodyElement)
        ..initialize();

      _videoPlayers[textureId] = player;
    } catch (e) {
      //no-op
    }
    addEventListener();
    return textureId;
  }

  void playerEventListener(String data) {
    Map<String, dynamic> dataMap = json.decode(data);
    int textureId = dataMap['textureId'] as int;
    if (_videoPlayers.keys.contains(textureId)) {
      _player(textureId).addEvent(dataMap);
    }
  }

  void addEventListener() {
    setProperty(
        html.window, 'onVdoPlayerEvent', js.allowInterop(playerEventListener));
  }

  @override
  Future<void> setLooping(int textureId, bool looping) async {}

  @override
  Future<void> play(int textureId) async {
    return promiseToFuture(playVideo(textureId));
  }

  @override
  Future<void> load(int textureId, String otp, String playbackInfo) async {
    return loadVideo(textureId, otp, playbackInfo);
  }

  @override
  Future<void> pause(int textureId) async {
    return promiseToFuture(pauseVideo(textureId));
  }

  @override
  Future<void> setVolume(int textureId, double volume) async {
    setPlayerVolume(textureId, volume);
  }

  @override
  Future<double> getVolume(int textureId) async {
    double volume = getPlayerVolume(textureId);
    return Future.value(volume);
  }

  @override
  Future<void> setMute(int textureId, bool mute) async {
    setPlayerMute(textureId, mute);
  }

  @override
  Future<void> updatePlayerConfigMap(int textureId, Map? configMap) async {
    if (configMap != null) {
      return promiseToFuture<dynamic>(
          updateConfigMap(textureId, json.encode(configMap)));
    }
  }

  @override
  Future<void> setPlaybackSpeed(int textureId, double speed) async {
    setPlaybackRate(textureId, speed);
  }

  @override
  Future<void> seekTo(int textureId, Duration position) async {
    videoSeekTo(textureId, position.inSeconds);
  }

  @override
  Future<Duration> getPosition(int textureId) async {
    double position = getVideoPosition(textureId);
    return Future.value(Duration(milliseconds: (position * 1000).toInt()));
  }

  @override
  Future<Duration> getBufferedPosition(int textureId) {
    double bufferedPosition = getVideoBufferedPosition(textureId);
    return Future.value(
        Duration(milliseconds: (bufferedPosition * 1000).toInt()));
  }

  @override
  Future<List<VideoTrack>> getVideoTracks(int textureId) async {
    /**
     * //Javascript object structure.
     * {id: 1, active: false, type: variant, bandwidth: 672000,
     * language: und,
     * label: 672 kbps, kind: null, width: 608, height: 1080, frameRate: 30,
     * pixelAspectRatio: 1:1, hdr: null, mimeType: video/mp4, audioMimeType: null,
     * videoMimeType: video/mp4, codecs: avc1.4d4028, audioCodec: null,
     * videoCodec: avc1.4d4028, primary: true, roles: [main], audioRoles: null,
     * forced: false, videoId: 2, audioId: null,
     * channelsCount: null, audioSamplingRate: null,
     * spatialAudio: false, tilesLayout: null, audioBandwidth: null,
     * videoBandwidth: 672000, originalVideoId: 1, originalAudioId: null,
     * originalTextId: null, originalImageId: null}
     */
    dynamic response =
        await promiseToFuture<dynamic>(getVideoQualities(textureId));
    Map<Object?, Object?> result = dartify(response) as Map<Object?, Object?>;
    List<VideoTrack> videoTracks = List.empty(growable: true);
    for (dynamic track in (result['qualities'] as List<dynamic>)) {
      Map<Object?, Object?> trackMap = track as Map<Object?, Object?>;
      VideoTrack videoTrack = VideoTrack(
          id: trackMap['id']! as int,
          bitrate: trackMap['bandwidth']! as int,
          width: trackMap['width']! as int,
          height: trackMap['height']! as int);
      videoTracks.add(videoTrack);
    }
    return Future.value(videoTracks);
  }

  @override
  Future<List<SubtitleTrack>> getSubtitleTracks(int textureId) async {
    /**
     * {visible: true, languages: [{id: 0, isDefault: true, lang: en, label: English, captionName: en,
     * url: https://d1z78r8i505acl.cloudfront.net/subtitle/C98NUuY0FDGNO.xm2kxpzd2v.vtt, active: true}, {id: 1, isDefault: false, lang: hi,
     * label: Hindi, captionName: hi, url: https://d1z78r8i505acl.cloudfront.net/subtitle/C98NUuY0FDGNO.4okqc0cvjv.vtt, active: false}, {id: 2, isDefault: false, lang: ur, label: Urdu, captionName: ur, url: https://d1z78r8i505acl.cloudfront.net/subtitle/C98NUuY0FDGNO.widpbhq0h7.vtt, active: false}, {id: 3, isDefault: null, lang: en, label: English, captionName: en-1, url: https://d1z78r8i505acl.cloudfront.net/subtitle/C98NUuY0FDGNO.4emojeenir.vtt, active: false}]}
     */
    dynamic response =
        await promiseToFuture<dynamic>(getCaptionLanguages(textureId));
    Map<Object?, Object?> result = dartify(response) as Map<Object?, Object?>;
    List<SubtitleTrack> subtitleTracks = List.empty(growable: true);
    for (dynamic track in (result['languages'] as List<dynamic>)) {
      Map<Object?, Object?> trackMap = track as Map<Object?, Object?>;
      SubtitleTrack subtitleTrack = SubtitleTrack(
          id: trackMap['id']! as int, language: trackMap['lang']! as String);
      subtitleTracks.add(subtitleTrack);
    }
    return Future.value(subtitleTracks);
  }

  @override
  Future<Duration> getDuration(int textureId) {
    double duration = getVideoDuration(textureId);
    return Future.value(Duration(milliseconds: (duration * 1000).toInt()));
  }

  @override
  Future<int> getPlaybackProperty(int textureId, String propertyName) async {
    dynamic response = await promiseToFuture<dynamic>(
        fetchPlaybackProperty(textureId, propertyName));
    int result = dartify(response) as int;
    return Future.value(result);
  }

  @override
  Future<MediaInfo> getMediaInfo(int textureId, String mediaId) async {
    dynamic response = await promiseToFuture<dynamic>(getMetaData(textureId));
    Object result = dartify(response) as Object;
    LinkedHashMap<Object?, Object?> metaDataMap =
        result as LinkedHashMap<Object?, Object?>;
    String? title = metaDataMap['title'] as String?;
    String? description = metaDataMap['description'] as String?;
    int? length = metaDataMap['duration'] as int?;
    Duration duration = Duration(seconds: length ?? 0);
    MediaInfo? mediaInfo = MediaInfo(
        mediaId: mediaId,
        type: MediaInfoType.streaming,
        title: title,
        description: description,
        duration: duration);
    return Future.value(mediaInfo);
  }

  @override
  Future<void> setAdaptive(int textureId) {
    return promiseToFuture<void>(setAdaptivePlayback(textureId));
  }

  @override
  Future<void> setVideoTrackId(int textureId, int videoTrackId) {
    return promiseToFuture<void>(setVideoQuality(textureId, videoTrackId));
  }

  @override
  Future<void> setTextTrackLanguage(int textureId, String? language) {
    if (language != null) {
      return promiseToFuture<void>(setTextTrack(textureId, language));
    } else {
      return promiseToFuture<void>(disableSubtitle(textureId));
    }
  }

  @override
  Future<VideoTrack?> getVideoTrack(int textureId) async {
    dynamic response =
        await promiseToFuture<dynamic>(getVideoQualities(textureId));
    Map<Object?, Object?> result = dartify(response) as Map<Object?, Object?>;
    VideoTrack? videoTrack;
    for (dynamic track in (result['qualities'] as List<dynamic>)) {
      Map<Object?, Object?> trackMap = track as Map<Object?, Object?>;
      if (trackMap['active'] as bool) {
        videoTrack = VideoTrack(
            id: trackMap['id']! as int,
            bitrate: trackMap['bandwidth']! as int,
            width: trackMap['width']! as int,
            height: trackMap['height']! as int);
      }
    }
    return Future.value(videoTrack);
  }

  @override
  Future<SubtitleTrack?> getSubtitleTrack(int textureId) async {
    dynamic response =
        await promiseToFuture<dynamic>(getCaptionLanguages(textureId));
    Map<Object?, Object?> result = dartify(response) as Map<Object?, Object?>;
    SubtitleTrack? subtitleTrack;
    for (dynamic track in (result['languages'] as List<dynamic>)) {
      Map<Object?, Object?> trackMap = track as Map<Object?, Object?>;
      if (trackMap['active'] as bool) {
        subtitleTrack = SubtitleTrack(
            id: trackMap['id']! as int, language: trackMap['lang']! as String);
      }
    }
    return Future.value(subtitleTrack);
  }

  // Retrieves a [VideoPlayer] by its internal `id`.
  // It must have been created earlier from the [create] method.
  VideoPlayer _player(int id) {
    return _videoPlayers[id]!;
  }

  @override
  Widget buildView(int textureId) {
    return HtmlElementView(viewType: 'vdoCipherPlayer-$textureId');
  }

  /// Sets the audio mode to mix with other sources (ignored)
  @override
  Future<void> setMixWithOthers(bool mixWithOthers) => Future<void>.value();

  @override
  Stream<VideoEvent> videoEventsFor(int textureId) {
    return _player(textureId).events;
  }
}
