import 'dart:async';
import 'dart:html' as html;
import 'package:vdocipher_flutter/src/web/vdo_player_platform_interface.dart';

/// Wraps a [html.VideoElement] so its API complies with what is expected by the plugin.
class VideoPlayer {
  /// Create a [VideoPlayer] from a [html.VideoElement] instance.
  VideoPlayer({required this.videoElement});

  html.HtmlElement videoElement;
  final StreamController<VideoEvent> _eventController =
      StreamController<VideoEvent>();

  Stream<VideoEvent> get events => _eventController.stream;

  void initialize() {
    // no-op

    _sendInitialized();
  }

  // Sends an [VideoEventType.initialized] [VideoEvent] with info about the wrapped video.
  void _sendInitialized() {
    _eventController.add(VideoEvent(eventType: VideoEventType.initialized));
  }

  void dispose() {
    videoElement.setInnerHtml("");
  }

  void addEvent(dynamic eventData) {
    VideoEventType videoEventType = VideoEventType.unknown;
    VideoEvent videoEvent = VideoEvent(eventType: videoEventType);
    switch (eventData['name']) {
      case 'initialized':
        videoEventType = VideoEventType.initialized;
        break;
      case 'apiReady':
        videoEventType = VideoEventType.apiReady;
        break;
      case 'waiting':
        videoEventType = VideoEventType.waiting;
        break;
      case 'playing':
        videoEventType = VideoEventType.playing;
        break;
      case 'play':
        videoEventType = VideoEventType.play;
        break;
      case 'buffered':
        videoEventType = VideoEventType.bufferingUpdate;
        if (eventData['value'] != null) {
          double value = eventData['value'] as double;
          Duration bufferedPosition =
              Duration(milliseconds: (value * 1000).toInt());
          videoEvent.copyWith(bufferedPosition: bufferedPosition);
        }
        break;
      case 'pause':
        videoEventType = VideoEventType.pause;
        break;
      case 'progress':
        videoEventType = VideoEventType.progress;
        if (eventData['value'] != null) {
          double value = eventData['value'] as double;
          Duration progress = Duration(milliseconds: (value * 1000).toInt());
          videoEvent.copyWith(progress: progress);
        }
        break;
      case 'ratechange':
        videoEventType = VideoEventType.ratechange;
        if (eventData['value'] != null) {
          double playbackRate = eventData['value'] as double;
          videoEvent.copyWith(playbackRate: playbackRate);
        }
        break;
      case 'ended':
        videoEventType = VideoEventType.ended;
        break;
      case 'error':
        videoEventType = VideoEventType.error;
        if (eventData['value'] != null) {
          String? message = eventData['value'] as String?;
          int code = eventData['code'] ?? 0;
          Error error = Error(message: message, code: code);
          videoEvent.copyWith(error: error);
        }
        break;
      case 'loadstart':
        videoEventType = VideoEventType.loadstart;
        break;
      case 'loadeddata':
        videoEventType = VideoEventType.loadeddata;
        if (eventData['value'] != null) {
          double value = eventData['value'] as double;
          Duration duration = Duration(milliseconds: (value * 1000).toInt());
          videoEvent.copyWith(duration: duration);
        }
        break;
      case 'videoQualityChange':
        videoEventType = VideoEventType.videoQualityChange;
        if (eventData['value'] != null) {
          int videoTrackId = eventData['value'] as int;
          videoEvent.copyWith(videoTrackId: videoTrackId);
        }
        break;
      case 'videoAdaptivenessChange':
        videoEventType = VideoEventType.videoAdaptivenessChange;
        if (eventData['value'] != null) {
          bool isAdaptive = eventData['value'] as bool;
          videoEvent.copyWith(isAdaptive: isAdaptive);
        }
        break;
      case 'captionLanguageChange':
        videoEventType = VideoEventType.captionLanguageChange;
        if (eventData['value'] != null) {
          int textTrackId = eventData['value'] as int;
          videoEvent.copyWith(textTrackId: textTrackId);
        }
        break;
      case 'volumechange':
        videoEventType = VideoEventType.volumechange;
        if (eventData['muted'] != null) {
          bool muted = eventData['muted'] as bool;
          videoEvent.copyWith(isMuted: muted);
        }
        if (eventData['value'] != null) {
          double volume = eventData['value'] as double;
          videoEvent.copyWith(volume: volume);
        }
        break;
      case 'captionVisibilityChange':
        videoEventType = VideoEventType.captionVisibilityChange;
        if (eventData['value'] != null) {
          bool isCaptionEnabled = eventData['value'] as bool;
          videoEvent.copyWith(isCaptionEnabled: isCaptionEnabled);
        }
        break;
      default:
        videoEventType = VideoEventType.unknown;
    }

    videoEvent.copyWith(eventType: videoEventType);
    _eventController.add(videoEvent);
  }
}
